import importlib

from scrapy.utils.misc import load_object

from . import defaults


class Scheduler(object):
    """
    Scheduler

    """

    def __init__(self,
                 persist=False,
                 flush_on_start=False,
                 queue_table=defaults.SCHEDULER_QUEUE_TABLE,
                 queue_cls=defaults.SCHEDULER_QUEUE_CLASS,
                 dupefilter_table=defaults.SCHEDULER_DUPEFILTER_TABLE,
                 dupefilter_cls=defaults.SCHEDULER_DUPEFILTER_CLASS,
                 idle_before_close=0,
                 serializer=None):
        """
        Initialize scheduler.

        :param persist: whether to restore previous progress
        :param flush_on_start: whether to clear progress when starting the crawler
        :param queue_table: name of the queue table
        :param queue_cls: class of the queue
        :param dupefilter_table: name of the duplicate filter table
        :param dupefilter_cls: class of the duplicate filter
        :param idle_before_close: maximum timeout when idle
        :param serializer: serialization tool
        """
        self.df = None
        self.queue = None
        self.spider = None
        if idle_before_close < 0:
            raise TypeError("idle_before_close cannot be negative")

        self.persist = persist
        self.flush_on_start = flush_on_start
        self.queue_table = queue_table
        self.queue_cls = queue_cls
        self.dupefilter_cls = dupefilter_cls
        self.dupefilter_table = dupefilter_table
        self.idle_before_close = idle_before_close
        self.serializer = serializer
        self.stats = None

    def __len__(self):
        return len(self.queue)

    @classmethod
    def from_settings(cls, settings):
        kwargs = {
            'persist': settings.getbool('SCHEDULER_PERSIST'),
            'flush_on_start': settings.getbool('SCHEDULER_FLUSH_ON_START'),
            'idle_before_close': settings.getint('SCHEDULER_IDLE_BEFORE_CLOSE'),
        }

        optional = {
            'queue_table': 'SCHEDULER_QUEUE_TABLE',
            'queue_cls': 'SCHEDULER_QUEUE_CLASS',
            'dupefilter_table': 'SCHEDULER_DUPEFILTER_TABLE',
            'dupefilter_cls': 'SCHEDULER_DUPEFILTER_CLASS',
            'serializer': 'SCHEDULER_SERIALIZER',
        }
        for name, setting_name in optional.items():
            val = settings.get(setting_name)
            if val:
                kwargs[name] = val

        if isinstance(kwargs.get('serializer'), str):
            kwargs['serializer'] = importlib.import_module(kwargs['serializer'])

        return cls(**kwargs)

    @classmethod
    def from_crawler(cls, crawler):
        """
        Inject crawler.

        :param crawler: crawler
        :return: scheduler object
        """
        instance = cls.from_settings(crawler.settings)
        instance.stats = crawler.stats
        return instance

    def open(self, spider):
        try:
            self.queue = load_object(self.queue_cls)(
                spider=spider,
                table_name=self.queue_table % {'spider': spider.name},
                key='queue',
                serializer=self.serializer,
            )
        except TypeError as e:
            raise ValueError(f"Failed to instantiate queue class '{self.queue_cls}': {e}")

        self.df = load_object(self.dupefilter_cls).from_spider(spider)

        if self.flush_on_start:
            self.flush()
        if len(self.queue):
            spider.log(f"Resuming crawl ({len(self.queue)} requests scheduled)")

    def close(self, reason):
        if not self.persist:
            self.flush()

    def flush(self):
        self.df.clear()
        self.queue.clear()

    def enqueue_request(self, request):
        if not request.dont_filter and self.df.request_seen(request):
            self.df.log(request, self.spider)
            return False
        if self.stats:
            self.stats.inc_value('scheduler/enqueued/db', spider=self.spider)
        self.queue.push(request)
        return True

    def next_request(self):
        block_pop_timeout = self.idle_before_close
        request = self.queue.pop(block_pop_timeout)
        if request and self.stats:
            self.stats.inc_value('scheduler/dequeued/db', spider=self.spider)
        return request

    def has_pending_requests(self):
        return len(self) > 0



STATS_TABLE = '%(spider)s_stats'

SCHEDULER_QUEUE_TABLE = '%(spider)s_requests'
SCHEDULER_QUEUE_CLASS = 'scrapy_db.queue.PriorityQueue'

SCHEDULER_DUPEFILTER_TABLE = '%(spider)s_dupefilter'
SCHEDULER_DUPEFILTER_CLASS = 'scrapy_db.dupefilter.DBDupeFilter'

SCHEDULER_PERSIST = True
START_URLS_TABLE = '%(spider)s_start_urls'

# Maximum idle time if the queue is empty
MAX_IDLE_TIME = 0

# Timeout period if the queue is empty during scheduling
SCHEDULER_IDLE_BEFORE_CLOSE = 0



import time
from collections.abc import Iterable

from scrapy import signals, FormRequest
from scrapy.exceptions import DontCloseSpider
from scrapy.spiders import Spider, CrawlSpider

from . import defaults
from .db import DBModel
from .utils import TextColor, is_dict


class DBBase(object):
    """
    Base class
    """
    table_name = None
    db_batch_size = None

    db = None

    spider_idle_start_time = int(time.time())
    max_idle_time = None

    def start_requests(self):
        return self.next_requests()

    def setup_db(self, crawler=None):
        if self.db is not None:
            return

        if crawler is None:
            crawler = getattr(self, 'crawler', None)

        if crawler is None:
            raise ValueError("crawler is required")

        settings = crawler.settings

        if self.table_name is None:
            self.table_name = settings.get(
                'START_URLS_TABLE', defaults.START_URLS_TABLE,
            )

        self.table_name = self.table_name % {'spider': self.name}

        if not self.table_name.strip():
            raise ValueError("table_name must not be empty")

        if self.db_batch_size is None:
            self.db_batch_size = settings.getint('CONCURRENT_REQUESTS')

        try:
            self.db_batch_size = int(self.db_batch_size)
        except (TypeError, ValueError):
            raise ValueError("db_batch_size must be an integer")

        self.logger.info("Reading start URLs from table_name '%(table_name)s' "
                         "(batch size: %(db_batch_size)s)",
                         self.__dict__)

        self.db = DBModel.build_model_from_settings(crawler.settings, self.table_name, 'start_url')

        if self.max_idle_time is None:
            self.max_idle_time = settings.getint(
                "MAX_IDLE_TIME_BEFORE_CLOSE",
                defaults.MAX_IDLE_TIME
            )

        try:
            self.max_idle_time = int(self.max_idle_time)
        except (TypeError, ValueError):
            raise ValueError("max_idle_time must be an integer")

        crawler.signals.connect(self.spider_idle, signal=signals.spider_idle)

    def next_requests(self):
        found = 0
        datas = self.db.fetch_data(batch_size=self.db_batch_size)
        for data in datas:
            reqs = self.make_request_from_data(data)
            if isinstance(reqs, Iterable):
                for req in reqs:
                    yield req
                    found += 1
                    self.logger.info(f'start req url:{req.url}')
            elif reqs:
                yield reqs
                found += 1
            else:
                self.logger.debug(f"Request not made from data: {data}")

        if found:
            self.logger.debug(f"Read {found} requests from '{self.table_name}'")

    def make_request_from_data(self, data):
        d = data.start_url
        parameter = is_dict(d)
        if isinstance(parameter, str):
            self.logger.warning(f"{TextColor.WARNING}WARNING: String request is deprecated, please use JSON data format. \
                Detail information, please check https://github.com/rmax/scrapy-redis#features{TextColor.ENDC}")
            return FormRequest(d, dont_filter=True)

        if parameter.get('url', None) is None:
            self.logger.warning(f"{TextColor.WARNING}The data from db has no url key in push data{TextColor.ENDC}")
            return []

        url = parameter.pop("url")
        method = parameter.pop("method").upper() if "method" in parameter else "GET"
        metadata = parameter.pop("meta") if "meta" in parameter else {}

        return FormRequest(url, dont_filter=True, method=method, formdata=parameter, meta=metadata)

    def schedule_next_requests(self):
        for req in self.next_requests():
            self.crawler.engine.crawl(req, spider=self)

    def spider_idle(self):
        if self.db is not None and len(self.db) > 0:
            self.spider_idle_start_time = int(time.time())

        self.schedule_next_requests()

        idle_time = int(time.time()) - self.spider_idle_start_time
        if self.max_idle_time != 0 and idle_time >= self.max_idle_time:
            return
        raise DontCloseSpider


class DBSpider(DBBase, Spider):

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        obj = super(DBSpider, cls).from_crawler(crawler, *args, **kwargs)
        obj.setup_db(crawler)
        return obj


class DBCrawlSpider(DBBase, CrawlSpider):

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        obj = super(DBCrawlSpider, cls).from_crawler(crawler, *args, **kwargs)
        obj.setup_db(crawler)
        return obj



from scrapy.utils.request import request_from_dict

from scrapy_db.db import DBModel
from scrapy_db.utils import CustomPickle


class Base(object):
    """
    Basic queue class
    """

    def __init__(self, spider, table_name, key, serializer=None):
        """
        Initialize spider queue object

        :param spider: spider object
        :param table_name: table name
        :param key: current object key
        :param serializer: serialization
        """
        if serializer is None:
            serializer = CustomPickle
        if not hasattr(serializer, 'loads'):
            raise TypeError(f"serializer does not implement 'loads' function: {serializer}")
        if not hasattr(serializer, 'dumps'):
            raise TypeError(f"serializer does not implement 'dumps' function: {serializer}")
        self.db = DBModel.build_model_from_settings(spider.settings, table_name % {'spider': spider.name}, key)
        self.spider = spider
        self.serializer = serializer

    def _encode_request(self, request):
        """
        Encode request

        :param request: request object
        :return: encode result
        """
        obj = request.to_dict(spider=self.spider)
        return self.serializer.dumps(obj)

    def _decode_request(self, encoded_request):
        """
        Decode request

        :param encoded_request: request to be decoded
        :return: decoding result
        """
        obj = self.serializer.loads(encoded_request)
        return request_from_dict(obj, spider=self.spider)

    def __len__(self):
        return len(self.db)

    def push(self, request):
        self.db.push(**{'key_': self._encode_request(request)})

    def pop(self, timeout=0):
        raise NotImplementedError

    def clear(self):
        self.db.drop_table()


class FifoQueue(Base):

    def pop(self, timeout=0):
        result = self.db.pop(timeout, desc=False)
        if result:
            return self._decode_request(result.key_)


class PriorityQueue(Base):

    def push(self, request):
        data = self._encode_request(request)
        score = -request.priority
        self.db.push(**{'key_': data, 'score': score})

    def pop(self, timeout=0):
        result = self.db.pop_by_score(timeout)
        if result:
            return self._decode_request(result.key_)


class LifoQueue(Base):

    def pop(self, timeout=0):
        result = self.db.pop(timeout)
        if result:
            return self._decode_request(result.key_)



import logging
import time

from scrapy.dupefilters import BaseDupeFilter
from scrapy.utils.request import fingerprint

from . import defaults
from .db import DBModel


class DBDupeFilter(BaseDupeFilter):
    """
    Duplicate filter
    """

    logger = logging.getLogger(__name__)

    def __init__(self, table, debug=False):
        """
        Initialize

        :param table: Table name
        :param debug: DUPEFILTER_DEBUG, whether to print duplicate values
        """
        self.table: DBModel = table
        self.debug = debug
        self.log_dupes = True

    @classmethod
    def from_settings(cls, settings):
        """
        Create an instance through settings

        :param settings: Spider settings
        :return: Instance of the current class
        """
        key = defaults.SCHEDULER_DUPEFILTER_TABLE % {'spider': int(time.time())}
        table = DBModel.build_model_from_settings(settings, key, 'dupelifter')
        debug = settings.getbool('DUPEFILTER_DEBUG')
        return cls(table=table, debug=debug)

    @classmethod
    def from_crawler(cls, crawler):
        """
        Create an instance through crawler

        :param crawler: Crawler object
        :return: Instance of the current class
        """
        return cls.from_settings(crawler.settings)

    def request_seen(self, request):
        """
        Returns True if the request has already been made

        :param request: Request object
        :return: Whether the request has been made
        """
        # request_fingerprint remove warnings
        fp = fingerprint(request).hex()
        added = self.table.db.select().where(self.table.db.key_ == fp).count()
        self.table.push(**{'key_': fp})
        return added != 0

    @classmethod
    def from_spider(cls, spider):
        """
        Create an instance through spider

        :param spider: Spider object
        :return: Instance of the current class
        """
        settings = spider.settings
        key = settings.get("SCHEDULER_DUPEFILTER_KEY", defaults.SCHEDULER_DUPEFILTER_TABLE) % {'spider': spider.name}
        table = DBModel.build_model_from_settings(settings, key, 'dupelifter')
        debug = settings.getbool('DUPEFILTER_DEBUG')
        return cls(table, debug=debug)

    def close(self, reason=''):
        """
        Clear data during shutdown

        :param reason: Shutdown reason
        :return: None
        """
        self.clear()

    def clear(self):
        """
        Clear fingerprint data

        :return: None
        """
        self.table.drop_table()

    def log(self, request, spider):
        """
        Prints the request

        :param request: Request
        :param spider: Spider object
        :return: None
        """
        if self.debug:
            msg = "Filtered duplicate request: %(request)s"
            self.logger.debug(msg, {'request': request}, extra={'spider': spider})
        elif self.log_dupes:
            msg = ("Filtered duplicate request %(request)s"
                   " - no more duplicates will be shown"
                   " (see DUPEFILTER_DEBUG to show all duplicates)")
            self.logger.debug(msg, {'request': request}, extra={'spider': spider})
            self.log_dupes = False



import copy
import datetime
from abc import ABCMeta, abstractmethod

from peewee import DateTimeField, CharField, BigAutoField, Model, IntegerField, SQL, BooleanField
from playhouse.db_url import connect

from scrapy_db.utils import execute_with_timeout


class BaseDB(metaclass=ABCMeta):
    """
    Base class for database queues
    """

    @abstractmethod
    def push(self, **value):
        """
        Add an element to the queue

        :param value: The element to add to the queue
        :return: None
        """
        pass

    @abstractmethod
    def pop(self, timeout=0, desc=True):
        """
        Remove an element from the queue, support FIFO and LIFO

        :param timeout: Timeout parameter
        :param desc: FIFO or LIFO, default is FIFO
        :return: The element removed from the queue
        """
        pass

    @abstractmethod
    def __len__(self):
        """
        Get the number of elements in the queue

        :return: The number of elements in the queue
        """
        pass

    @abstractmethod
    def pop_by_score(self, timeout=0):
        """
        Remove an element from the queue by score

        :param timeout: Timeout parameter
        :return: The element removed from the queue
        """
        pass


# The default fields for model classes
_attributes = {'id': BigAutoField(primary_key=True),
               'create_time': DateTimeField(default=datetime.datetime.now,
                                            constraints=[SQL('DEFAULT CURRENT_TIMESTAMP')]),
               'update_time': DateTimeField(default=datetime.datetime.now,
                                            constraints=[SQL('ON UPDATE CURRENT_TIMESTAMP')]),
               'Meta': type('Meta', (object,), {'table_name': None, 'database': None})}


def get_model_class_for_db(name, attrs) -> Model:
    """
    Return a model class based on name and attributes

    :param name: The prefix of the class name
    :param attrs: The attributes of the model class
    :return: The model class
    """
    attrs_ = copy.deepcopy(_attributes)
    attrs_.update(attrs)
    return type(f'{name.capitalize()}Model', (Model,), attrs_)  # noqa


# A dictionary that stores the fields of the model class
_params = {
    'dupelifter': {
        'key_': CharField(),
    },
    'start_url': {
        'start_url': CharField(),
        'deleted': BooleanField(default=False)
    },
    'queue': {
        'key_': CharField(max_length=10000),
        'score': IntegerField(index=True),
        'deleted': BooleanField(default=False)
    }
}


def db_require(fun):
    """
    Check if db exists

    :param fun: The function to be executed
    :return: The inner function
    """

    def inner(self, *args, **kwargs):
        if self.db is None:
            raise ValueError('db is None!')
        return fun(self, *args, **kwargs)

    return inner


class DBModel(BaseDB):
    def __init__(self, db=None):
        self.db: Model = db

    @classmethod
    def build_model_from_settings(cls, settings, name, key):
        """
        Initialize an instance of the current class based on the settings and name

        :param settings: The settings
        :param name: The name
        :param key: The key in the field dictionary
        :return: An instance of the model class
        """
        attrs = _params.get(key)
        if not attrs:
            raise ValueError(f'key {key} not found in params!')
        model = get_model_class_for_db(name, attrs)
        url = settings.get('DB_URL')
        assert url is not None
        model._meta.database = connect(url)  # noqa
        model._meta.table_name = name  # noqa
        model.create_table()
        return cls(model)

    @db_require
    def push(self, **value):
        self.db.create(**value)

    @db_require
    def drop_table(self):
        self.db.drop_table()

    @execute_with_timeout
    @db_require
    def pop(self, timeout=0, desc=True, batch_size=None):
        if not batch_size:
            result = self.db.select().where(self.db.deleted == 0).order_by(
                self.db.id.desc() if desc else self.db.id.asc()).limit(1).first()
            if result:
                self._delete(result)
            return result
        else:
            result = self.db.select().where(self.db.deleted == 0).order_by(
                self.db.id.desc() if desc else self.db.id.asc()).limit(batch_size)
            if result.count() > 0:
                self._delete(*copy.deepcopy(list(result)))
            return result

    def _delete(self, *results):
        if results:
            self.db.update(deleted=1).where(self.db.id.in_([a.id for a in results])).execute()

    @db_require
    def fetch_data(self, batch_size=1):
        result = self.db.select().where(self.db.deleted == 0).order_by(self.db.id.desc()).limit(batch_size)
        if result.count() > 0:
            self._delete(*copy.deepcopy(list(result)))
        return list(result)

    @db_require
    def __len__(self):
        return self.db.select().where(self.db.deleted == 0).count()

    @execute_with_timeout
    @db_require
    def pop_by_score(self, timeout=0):
        result = self.db.select().where(self.db.deleted == 0).order_by(self.db.score.desc()).limit(1).first()
        if result:
            self._delete(result)
        return result if result else None



import base64
import pickle
import time
from ast import literal_eval


class TextColor:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def is_dict(content):
    try:
        return literal_eval(content)
    except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError):
        return content


class CustomPickle(object):
    @staticmethod
    def loads(data):
        return pickle.loads(base64.b64decode(data))

    @staticmethod
    def dumps(*args, **kwargs):
        return base64.b64encode(pickle.dumps(*args, **kwargs)).decode()


def execute_with_timeout(func):
    def inner(*args, **kwargs):
        timeout_ = kwargs.get('timeout', 0)

        if timeout_ == 0:
            return func(*args, **kwargs)
        t = time.time()
        while True:
            if time.time() - t >= timeout_:
                break
            if r := func(*args, **kwargs):
                return r

    return inner
